# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi
from nbt.nbt import NBTFile, TAG_List, TAG_Int, TAG_Short, TAG_Byte_Array, TAG_String


def extract_region_with_data(init_pos, end_pos):
    """
    Extract a Minecraft world region with the id and data of the blocks

    :return: bytearrays for blocks ids and block data
    """
    from mcthings.scene import Scene

    size_x = end_pos.x - init_pos.x + 1
    size_z = end_pos.z - init_pos.z + 1
    size_y = end_pos.y - init_pos.y + 1

    blocks_bytes = bytearray()
    data_bytes = bytearray()

    # Use the same loop than reading Schematic format: x -> z -> y
    for y in range(0, size_y):
        for z in range(0, size_z):
            for x in range(0, size_x):
                block_pos = mcpi.vec3.Vec3(init_pos.x + x, init_pos.y + y, init_pos.z + z)
                block = Scene.server.getBlockWithData(block_pos.x, block_pos.y, block_pos.z)
                blocks_bytes.append(block.id)
                data_bytes.append(block.data)

    return blocks_bytes, data_bytes


def build_schematic_nbt(init_pos, end_pos):
    """
    Creates a NBT Object with the schematic data

    :return: The NBT object with the Schematic
    """
    size_x = end_pos.x - init_pos.x + 1
    size_z = end_pos.z - init_pos.z + 1
    size_y = end_pos.y - init_pos.y + 1

    # Prepare the NBT Object
    nbtfile = NBTFile()
    nbtfile.name = "Schematic"
    nbtfile.tags.append(TAG_Short(name="Width", value=size_x))
    nbtfile.tags.append(TAG_Short(name="Height", value=size_y))
    nbtfile.tags.append(TAG_Short(name="Length", value=size_z))
    nbt_blocks = TAG_Byte_Array(name="Blocks")
    nbtfile.tags.append(nbt_blocks)
    nbt_data = TAG_Byte_Array(name="Data")
    nbtfile.tags.append(nbt_data)

    # Additional fields need
    nbtfile.tags.append(TAG_String(name="Materials", value="Alpha"))
    nbtfile.tags.append(TAG_Int(name="'WEOriginX'", value=0))
    nbtfile.tags.append(TAG_Int(name="'WEOriginY'", value=0))
    nbtfile.tags.append(TAG_Int(name="'WEOriginZ'", value=0))
    nbtfile.tags.append(TAG_Int(name="'WEOffsetX'", value=0))
    nbtfile.tags.append(TAG_Int(name="'WEOffsetY'", value=0))
    nbtfile.tags.append(TAG_Int(name="'WEOffsetZ'", value=0))
    entities_list = TAG_List(name="Entities", type=TAG_Int)
    nbtfile.tags.append(entities_list)
    tile_entities_list = TAG_List(name="TileEntities", type=TAG_Int)
    nbtfile.tags.append(tile_entities_list)

    # Collect all blocks ids and data
    (blocks_bytes, data_bytes) = extract_region_with_data(init_pos, end_pos)

    nbt_blocks.value = blocks_bytes
    nbt_data.value = data_bytes

    return nbtfile



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

# TODO: at some point this must be a real Singleton
import pickle

from mcpi.vec3 import Vec3

from mcthings.utils import build_schematic_nbt


class Scene:
    """
    A scene is a container for all the things built using McThings.
    A scene can be built, unbuilt and moved. There is only one scene
    in a program using McThings. Things built are added automatically to
    the Scene. A Scene can also be loaded from a file, and
    it can be saved to a file.

    Before adding Things to the Scene, it must be connected to a
    Minecraft server (fill the Scene.server attribute)
    """

    things = []
    """ map with the things in the scene"""
    server = None
    """ Minecraft server in which create things"""
    _position = None

    @classmethod
    def add(cls, thing):
        """ Add a new thing to the scene """
        if not cls.things:
            # The initial position of the scene is the position
            # of its first thing added
            cls._position = thing.position
        cls.things.append(thing)

    @classmethod
    def build(cls):
        """ Build all the things inside the Scene """
        for thing in cls.things:
            thing.build()

    @classmethod
    def unbuild(cls):
        """ Unbuild all the things inside the Scene """
        for thing in cls.things:
            thing.unbuild()

    @classmethod
    def reposition(cls, position):
        """
        Move all the things in the scene to a new relative position

        :param position: new position for the Scene
        :return:
        """

        # All the things inside the scene must be moved
        diff_x = position.x - cls._position.x
        diff_y = position.y - cls._position.y
        diff_z = position.z - cls._position.z

        for thing in cls.things:
            repos_x = thing.position.x + diff_x
            repos_y = thing.position.y + diff_y
            repos_z = thing.position.z + diff_z

            thing._position = (Vec3(repos_x, repos_y, repos_z))

    @classmethod
    def move(cls, position):
        """
        Move the scene to a new position

        :param position: new position
        :return:
        """

        cls.unbuild()
        cls.reposition(position)
        cls.build()

    @classmethod
    def load(cls, file_path):
        """ Load a scene from a file (but no build it yet) """
        Scene.things = pickle.load(open(file_path, "rb"))
        if Scene.things:
            Scene._position = Scene.things[0].position

    @classmethod
    def save(cls, file_path):
        """ Save a scene to a file """
        pickle.dump(Scene.things, open(file_path, "wb"))

    @classmethod
    def to_schematic(cls, file_path):
        """
        Save the Scene into a Schematic file

        :file_path: file in which to export the Scene in Schematic format
        :return: the Schematic object
        """

        # Default init values
        init_pos = Vec3(cls._position.x, cls._position.y, cls._position.z)
        end_pos = Vec3(cls._position.x, cls._position.y, cls._position.z)

        for thing in cls.things:
            ip = thing.position
            ep = thing.end_position

            # Update init position
            init_pos.x = ip.x if ip.x < init_pos.x else init_pos.x
            init_pos.y = ip.y if ip.y < init_pos.y else init_pos.y
            init_pos.z = ip.z if ip.z < init_pos.z else init_pos.z

            # Update end position
            end_pos.x = ep.x if ep.x > end_pos.x else end_pos.x
            end_pos.y = ep.y if ep.y > end_pos.y else end_pos.y
            end_pos.z = ep.z if ep.z > end_pos.z else end_pos.z

            build_schematic_nbt(init_pos, end_pos).write_file(file_path)



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

from mcthings.scene import Scene
from mcthings.thing import Thing


class Sphere(Thing):

    radius = 5
    """ radius of the Sphere """
    height = 0

    def build(self):
        Scene.drawing.drawSphere(
            self.position.x,
            self.position.y + self.height - 1,
            self.position.z,
            self.radius,
            self.block)
        self._end_position = self.position


class SphereHollow(Thing):

    radius = None
    """ radius of the Hollow Sphere """
    height = 0

    def build(self):
        Scene.drawing.drawHollowSphere(
            self.position.x,
            self.position.y + self.height - 1,
            self.position.z,
            self.radius,
            self.block)
        self._end_position = self.position




#!/usr/bin/env python3

# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import codecs
import os
import re
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
readme_md = os.path.join(here, 'README.md')

version_py = os.path.join(here, 'mcthings', '_version.py')
with codecs.open(version_py, 'r', encoding='utf-8') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

# Get the package description from the README.md file
with codecs.open(readme_md, encoding='utf-8') as f:
    long_description = f.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='mcthings',
    version=version,
    packages=['mcthings'],
    include_package_data=True,
    license='ASL',
    description='McThings is a python library for building things in Minecraft',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/juntosdesdecasa/mcthings',
    author='Alvaro del Castillo',
    author_email='alvaro.delcastillo@gmail.com',
    keywords="development library minecraft buildings games",
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development'
    ],
    install_requires=[
        'mcpi',
        'minecraftstuff',
        'nbt'
    ],
    python_requires='>=3.4'
)



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

from mcthings.scene import Scene
from mcthings.thing import Thing


class Circle(Thing):

    radius = None
    """ radius of the Sphere """

    def build(self):
        Scene.drawing.drawCircle(
            self.position.x,
            self.position.y,
            self.position.z,
            self.radius,
            self.block)
        self._end_position = self.position
